import logging
from collections import defaultdict, namedtuple

from clint.textui import colored, indent, puts

from .utils import fullpath, bold

logger = logging.getLogger(__name__)


Job = namedtuple('Job', 'folder path filters actions')
Job.__doc__ = """
    :param str folder:   the folder the file was found in
    :param Path path:    the path of the file to handle
    :param list filters: the filters that apply to the path
    :param list actions: the actions which should be executed
"""


def all_files_for_rule(rule):
    for folder in rule.folders:
        globstr = '**/*.*' if rule.subfolders else '*.*'
        for path in fullpath(folder).glob(globstr):
            if path.is_file() and (
                    rule.system_files or
                    path.name not in ('thumbs.db', 'desktop.ini', '.DS_Store')):
                yield (folder, path)


def find_jobs(rules):
    for rule in rules:
        for folder, path in all_files_for_rule(rule):
            if all(f.matches(path) for f in rule.filters):
                yield Job(
                    folder=folder, path=path, filters=rule.filters,
                    actions=rule.actions)


def sort_by_folder(jobs):
    result = defaultdict(list)
    for job in jobs:
        result[job.folder].append(job)
    return result


def filter_pipeline(job):
    result = {}
    for filter_ in job.filters:
        result.update(filter_.parse(job.path))
    return result


def action_pipeline(job: Job, attrs: dict, simulate: bool):
    try:
        current_path = job.path.resolve()
        for action in job.actions:
            new_path = action.run(
                basedir=fullpath(job.folder),
                path=current_path,
                attrs=attrs,
                simulate=simulate)
            if new_path is not None:
                current_path = new_path
    except Exception as e:
        logger.exception(e)
        action.print('%s %s' % (colored.red('ERROR!', bold=True), e))


def execute_rules(rules, simulate):
    jobs = list(find_jobs(rules))
    if not jobs:
        msg = 'Nothing to do.'
        logger.info(msg)
        puts(msg)
        return

    jobs_by_folder = sort_by_folder(jobs)
    first = True
    for folder, jobs in sorted(jobs_by_folder.items()):
        # newline between folders
        if not first:
            puts()
        first = False

        puts('Folder %s:' % bold(folder))
        with indent(2):
            for job in jobs:
                try:
                    folderpath = fullpath(job.folder)
                    relative_path = str(job.path.relative_to(folderpath))
                except ValueError:
                    relative_path = job.path.name
                puts('File %s:' % bold(relative_path))
                with indent(2):
                    attrs = filter_pipeline(job)
                    action_pipeline(job=job, attrs=attrs, simulate=simulate)



import logging
import os
import shutil

from organize.utils import Path, fullpath, find_unused_filename

from .action import Action
from .trash import Trash


class Move(Action):

    """
    Move a file to a new location. The file can also be renamed.
    If the specified path does not exist it will be created.

    If you only want to rename the file and keep the folder, it is
    easier to use the Rename-Action.

    :param str dest:
        The destination folder or path.
        If `dest` ends with a slash / backslash, the file will be moved into
        this folder and not renamed.

    :param bool overwrite:
        specifies whether existing files should be overwritten.
        Otherwise it will start enumerating files (append a counter to the
        filename) to resolve naming conflicts. [Default: False]

    Examples:
        - Move all pdfs and jpgs from the desktop into the folder "~/Desktop/media/".
          Filenames are not changed.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - Extension:
                      - pdf
                      - jpg
                actions:
                  - Move: '~/Desktop/media/'

        - Use a placeholder to move all .pdf files into a "PDF" folder and all
          .jpg files into a "JPG" folder. Existing files will be overwritten.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - Extension:
                      - pdf
                      - jpg
                actions:
                  - Move:
                      dest: '~/Desktop/{extension.upper}/'
                      overwrite: true

        - Move pdfs into the folder `Invoices`. Keep the filename but do not
          overwrite existing files (adds an index to the file)

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop/Invoices
                filters:
                  - Extension:
                      - pdf
                actions:
                  - Copy:
                      dest: '~/Documents/Invoices/'
                      overwrite: false
    """

    def __init__(self, dest: str, overwrite=False):
        self.dest = dest
        self.overwrite = overwrite
        self.log = logging.getLogger(__name__)

    def run(self, basedir: Path, path: Path, attrs: dict, simulate: bool):
        full_path = fullpath(path)

        expanded_dest = self.fill_template_tags(
            self.dest, basedir, path, attrs)
        # if only a folder path is given we append the filename to have the full
        # path. We use os.path for that because pathlib removes trailing slashes
        if expanded_dest.endswith(('\\', '/')):
            expanded_dest = os.path.join(expanded_dest, path.name)

        new_path = fullpath(expanded_dest)
        new_path_exists = new_path.exists()
        new_path_samefile = new_path_exists and new_path.samefile(full_path)
        if new_path_exists and not new_path_samefile:
            if self.overwrite:
                self.print('File already exists')
                Trash().run(basedir, path=new_path, attrs=attrs, simulate=simulate)
            else:
                new_path = find_unused_filename(new_path)

        if new_path_samefile and new_path == full_path:
            self.print('Keep location')
        else:
            self.print('Move to "%s"' % new_path)
            if not simulate:
                self.log.info(
                    'Creating folder if not exists: %s', new_path.parent)
                new_path.parent.mkdir(parents=True, exist_ok=True)
                self.log.info('Moving "%s" to "%s"', full_path, new_path)
                shutil.move(src=str(full_path), dst=str(new_path))
        return new_path

    def __str__(self):
        return 'Move(dest=%s, overwrite=%s)' % (self.dest, self.overwrite)



from typing import Optional
from clint.textui import puts
from organize.utils import Path


class Error(Exception):
    pass


class TemplateAttributeError(Error):
    pass


class Action:

    def run(self, basedir: Path, path: Path, attrs: dict, simulate: bool) -> Optional[Path]:
        # if you change the file path, return the new path here
        raise NotImplementedError

    def print(self, msg):
        """ print a message for the user """
        puts('- [%s] %s' % (self.__class__.__name__, msg))

    @staticmethod
    def fill_template_tags(msg: str, basedir:Path, path: Path, attrs: dict) -> str:
        try:
            return msg.format(basedir=basedir, path=path, **attrs)
        except AttributeError as exc:
            cause = exc.args[0]
            raise TemplateAttributeError(
                'Missing template variable %s for "%s"' % (cause, msg))

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return '<%s>' % str(self)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.__dict__ == other.__dict__)



import logging
from collections import namedtuple
from typing import List

import yaml

from . import actions, filters
from .utils import Path, first_key, flatten

logger = logging.getLogger(__name__)
Rule = namedtuple('Rule', 'filters actions folders subfolders system_files')


class Config:

    def __init__(self, config: dict):  # type: (dict)
        self.config = config

    @classmethod
    def parse_yaml(cls, config: str) -> dict:
        try:
            return yaml.load(config)
        except yaml.YAMLError as e:
            raise cls.ParsingError(e)

    @classmethod
    def from_string(cls, config: str) -> 'Config':
        return cls(cls.parse_yaml(config))

    @classmethod
    def from_file(cls, path: Path) -> 'Config':
        with path.open() as f:
            return cls.from_string(f.read())

    def yaml(self) -> str:
        if 'rules' not in self.config:
            raise self.NoRulesFoundError()
        data = {'rules': self.config['rules']}
        yaml.Dumper.ignore_aliases = lambda self, data: True
        return yaml.dump(data, default_flow_style=False, default_style="'")

    @staticmethod
    def parse_folders(rule_item):
        # the folder list is flattened so we can use encapsulated list
        # definitions in the config file.
        yield from flatten(rule_item['folders'])

    def _get_filter_class_by_name(self, name):
        try:
            return getattr(filters, name)
        except AttributeError as e:
            raise self.Error('%s is no valid filter' % name) from e

    def _get_action_class_by_name(self, name):
        try:
            return getattr(actions, name)
        except AttributeError as e:
            raise self.Error('%s is no valid action' % name) from e

    @staticmethod
    def _class_instance_with_args(Cls, args):
        if args is None:
            return Cls()
        elif isinstance(args, list):
            return Cls(*args)
        elif isinstance(args, dict):
            return Cls(**args)
        return Cls(args)

    def instantiate_filters(self, rule_item):
        # filter list can be empty
        try:
            filter_list = rule_item['filters']
        except KeyError:
            return
        if not filter_list:
            return
        if not isinstance(filter_list, list):
            raise self.FiltersNoListError()

        for filter_item in filter_list:
            # filter with arguments
            if filter_item is None:
                yield None
            elif isinstance(filter_item, dict):
                name = first_key(filter_item)
                args = filter_item[name]
                filter_class = self._get_filter_class_by_name(name)
                yield self._class_instance_with_args(filter_class, args)
            # only given filter name without args
            elif isinstance(filter_item, str):
                name = filter_item
                filter_class = self._get_filter_class_by_name(name)
                yield filter_class()
            else:
                raise self.Error('Unknown filter: %s' % filter_item)

    def instantiate_actions(self, rule_item):
        action_list = rule_item['actions']
        if not isinstance(action_list, list):
            raise self.ActionsNoListError()

        for action_item in action_list:
            if isinstance(action_item, dict):
                name = first_key(action_item)
                args = action_item[name]
                action_class = self._get_action_class_by_name(name)
                yield self._class_instance_with_args(action_class, args)
            elif isinstance(action_item, str):
                name = action_item
                action_class = self._get_action_class_by_name(name)
                yield action_class()
            else:
                raise self.Error('Unknown action: %s' % action_item)

    @property
    def rules(self) -> List[Rule]:
        """ :returns: A list of instantiated Rules """
        if 'rules' not in self.config:
            raise self.NoRulesFoundError()
        result = []
        for i, rule_item in enumerate(self.config['rules']):
            # skip disabled rules
            if not rule_item.get('enabled', True):
                continue

            rule_folders = list(self.parse_folders(rule_item))
            rule_filters = list(self.instantiate_filters(rule_item))
            rule_actions = list(self.instantiate_actions(rule_item))

            if not rule_folders:
                logger.warning('No folders given for rule %s!', i + 1)
            if not rule_filters:
                logger.warning('No filters given for rule %s!', i + 1)
            if not rule_actions:
                logger.warning('No actions given for rule %s!', i + 1)

            rule = Rule(
                folders=rule_folders,
                filters=rule_filters,
                actions=rule_actions,
                subfolders=rule_item.get('subfolders', False),
                system_files=rule_item.get('system_files', False),
            )
            result.append(rule)
        return result

    class Error(Exception):
        pass

    class NoRulesFoundError(Error):
        def __str__(self):
            return 'No rules found in configuration file'

    class ParsingError(Error):
        pass

    class NoFoldersFoundError(Error):
        pass

    class NoFiltersFoundError(Error):
        pass

    class NoActionsFoundError(Error):
        pass

    class FiltersNoListError(Error):
        def __str__(self):
            return 'Please specify your filters as a YAML list'

    class ActionsNoListError(Error):
        def __str__(self):
            return 'Please specify your actions as a YAML list'



from .filter import Filter


class Filename(Filter):

    """
    Match files by filename

    :param str startswith:
        The filename must begin with the given string

    :param str contains:
        The filename must contain the given string

    :param str endswith:
        The filename (without extension) must end with the given string

    :param bool case_sensitive = True:
        By default, the matching is case sensitive. Change this to False to use
        case insensitive matching.

    Examples:
        - Match all files starting with 'Invoice':

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - Filename:
                      startswith: Invoice
                actions:
                  - Echo: 'This is an invoice'

        - Match all files starting with 'A' end containing the string 'hole'
          (case insensitive)

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - Filename:
                      startswith: A
                      contains: hole
                      case_sensitive: false
                actions:
                  - Echo: 'Found a match.'
    """

    def __init__(self, startswith='', contains='', endswith='',
                 case_sensitive=True):
        self.startswith = startswith
        self.contains = contains
        self.endswith = endswith
        self.case_sensitive = case_sensitive

        if not self.case_sensitive:
            self.startswith = self.startswith.lower()
            self.contains = self.contains.lower()
            self.endswith = self.endswith.lower()

    def matches(self, path):
        filename = self._filename(path)
        return (
            filename.startswith(self.startswith) and
            filename.endswith(self.endswith) and
            self.contains in filename)

    def _filename(self, path):
        if not self.case_sensitive:
            return path.stem.lower()
        else:
            return path.stem



__title__ = 'organize-tool'
__description__ = 'The file management automation tool.'
__url__ = 'http://github.com/tfeldmann/organize'
__version__ = '1.2'
__author__ = 'Thomas Feldmann'
__author_email__ = 'mail@tfeldmann.de'
__license__ = 'MIT'
__copyright__ = 'Copyright (c)Thomas Feldmann'



import re
from organize.utils import DotDict
from .filter import Filter


class Regex(Filter):

    """
    Matches filenames with the given regular expression

    :param str expr:
        The regular expression to be matched.

    Any named groups in your regular expression will be returned like this:

    :returns:
        - ``{regex.yourgroupname}`` -- The text matched with the named group
          ``(?P<yourgroupname>)``

    Examples:
        - Match an invoice with a regular expression:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - Regex: '^RG(\d{12})-sig\.pdf$'
                actions:
                  - Move: '~/Documents/Invoices/1und1/'

        - Match and extract data from filenames with regex named groups:
          This is just like the previous example but we rename the invoice using
          the invoice number extracted via the regular expression and the named
          group ``the_number``.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - Regex: '^RG(?P<the_number>\d{12})-sig\.pdf$'
                actions:
                  - Move: ~/Documents/Invoices/1und1/{regex.the_number}.pdf
    """

    def __init__(self, expr):
        self.expr = re.compile(expr)

    def matches(self, path):
        return bool(self.expr.match(path.name))

    def parse(self, path):
        result = DotDict(self.expr.match(path.name).groupdict())
        return {'regex': result}



import logging
from organize.utils import Path
from .action import Action


class Echo(Action):

    """
    Prints the given (formatted) message. This can be useful to test your rules,
    especially if you use formatted messages.

    :param str msg: The message to print (can be formatted)

    Example:
        - Prints "Found old file" for each file older than one year:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - LastModified:
                      days: 365
                actions:
                  - Echo: 'Found old file'

        - Prints "Hello World!" and filepath for each file on the desktop:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders:
                  - ~/Desktop
                actions:
                  - Echo: 'Hello World! {path}'

        - This will print something like ``Found a PNG: "test.png"`` for each
          file on your desktop:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders:
                  - ~/Desktop
                filters:
                  - Extension
                actions:
                  - Echo: 'Found a {extension.upper}: "{path.name}"'

        - Show the ``{basedir}`` and ``{path}`` of all files in '~/Downloads',
          '~/Desktop' and their subfolders:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders:
                  - ~/Desktop
                  - ~/Downloads
                subfolders: true
                actions:
                  - Echo: 'Basedir: {basedir}'
                  - Echo: 'Path:    {path}'
    """

    def __init__(self, msg):
        self.msg = msg
        self.log = logging.getLogger(__name__)

    def run(self, basedir: Path, path: Path, attrs: dict, simulate: bool):
        self.log.debug(
            'Echo msg "%s" for path: "%s" with attrs: "%s"',
            self.msg, path, attrs)
        full_msg = self.fill_template_tags(self.msg, basedir, path, attrs)
        self.log.info('Console output: %s', full_msg)
        self.print('%s' % full_msg)

    def __str__(self):
        return 'Echo(msg="%s")' % self.msg



import sys
from collections import OrderedDict

import colorama

# in python < 3.6 the pathlib module misses some features so we have to import
# a backported alternative
if sys.version_info < (3, 6):
    from pathlib2 import Path
else:
    from pathlib import Path


def fullpath(path):
    """ Expand '~' and resolve the given path """
    return Path(path).expanduser().resolve(strict=False)


def bold(text):
    # inspired by a feature request from the clint library
    # https://github.com/kennethreitz/clint/issues/157
    if sys.stdout.isatty():
        return ''.join([colorama.Style.BRIGHT, text, colorama.Style.NORMAL])
    return text


def flatten(arr):
    if arr == []:
        return []
    if not isinstance(arr, list):
        return [arr]
    return flatten(arr[0]) + flatten(arr[1:])


def first_key(dic: dict):
    return list(dic.keys())[0]


class DotDict(OrderedDict):

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        if not key.startswith('_OrderedDict__'):
            self[key] = value
        else:
            OrderedDict.__setattr__(self, key, value)

    def __delattr__(self, key):
        try:
            self.pop(key)
        except KeyError:
            OrderedDict.__delattr__(self, key)

    def __eq__(self, other):
        return dict.__eq__(self, other)

    def __str__(self):
        return '{%s}' % ', '.join('%r: %r' % (key, self[key]) for key in self)


def find_unused_filename(path: Path) -> Path:
    """
    we assume path already exists. This function then adds a counter to the
    filename until we find a unused filename.
    """
    stem = path.stem
    try:
        splitstem = stem.split(' ')
        if len(splitstem) < 2:
            raise ValueError()
        count = int(splitstem[-1])
        stem = ' '.join(splitstem[:-1])
    except (ValueError, IndexError):
        count = 1
    while True:
        count += 1
        tmp_path = path.with_name('%s %s%s' % (stem, count, path.suffix))
        if not tmp_path.exists():
            return tmp_path



import logging
import os
import shutil

from organize.utils import Path, fullpath, find_unused_filename

from .action import Action
from .trash import Trash


class Copy(Action):

    """
    Copy a file to a new location.
    If the specified path does not exist it will be created.

    :param str dest:
        The destination where the file should be copied to.
        If `dest` ends with a slash / backslash, the file will be copied into
        this folder and keep its original name.

    :param bool overwrite:
        specifies whether existing files should be overwritten.
        Otherwise it will start enumerating files (append a counter to the
        filename) to resolve naming conflicts. [Default: False]

    Examples:
        - Copy all pdfs into `~/Desktop/somefolder/` and keep filenames

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - Extension: pdf
                actions:
                  - Copy: '~/Desktop/somefolder/'

        - Use a placeholder to copy all .pdf files into a "PDF" folder and all .jpg
          files into a "JPG" folder. Existing files will be overwritten.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop
                filters:
                  - Extension:
                      - pdf
                      - jpg
                actions:
                  - Copy:
                      dest: '~/Desktop/{extension.upper}/'
                      overwrite: true

        - Copy into the folder `Invoices`. Keep the filename but do not
          overwrite existing files (adds an index to the file)

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: ~/Desktop/Invoices
                filters:
                  - Extension:
                      - pdf
                actions:
                  - Copy:
                      dest: '~/Documents/Invoices/'
                      overwrite: false
    """

    def __init__(self, dest: str, overwrite=False):
        self.dest = dest
        self.overwrite = overwrite
        self.log = logging.getLogger(__name__)

    def run(self, basedir: Path, path: Path, attrs: dict, simulate: bool):
        full_path = fullpath(path)

        expanded_dest = self.fill_template_tags(self.dest, basedir, path, attrs)
        # if only a folder path is given we append the filename to have the full
        # path. We use os.path for that because pathlib removes trailing slashes
        if expanded_dest.endswith(('\\', '/')):
            expanded_dest = os.path.join(expanded_dest, path.name)

        new_path = fullpath(expanded_dest)
        if new_path.exists() and not new_path.samefile(full_path):
            if self.overwrite:
                self.print('File already exists')
                Trash().run(basedir, path=new_path, attrs=attrs, simulate=simulate)
            else:
                new_path = find_unused_filename(new_path)

        self.print('Copy to "%s"' % new_path)
        if not simulate:
            self.log.info('Creating folder if not exists: %s', new_path.parent)
            new_path.parent.mkdir(parents=True, exist_ok=True)
            self.log.info('Copying "%s" to "%s"', full_path, new_path)
            shutil.copy2(src=str(full_path), dst=str(new_path))

        # the next actions should handle the original file
        return None

    def __str__(self):
        return 'Copy(dest=%s, overwrite=%s)' % (self.dest, self.overwrite)
